/**
 * By Shujie Han
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This lib is used to parse MS snapshots (http://iotta.snia.org/tracetypes/6).
 */

#include "libhashfile.h"
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/utsname.h>
#include <time.h>
#include <unistd.h>
#define MAX_B 255

char *get_OS(struct hashfile_handle *handle) { return handle->metadata.OS; }
char *get_hostname(struct hashfile_handle *handle) {
  return handle->metadata.hostname;
}
char *get_time(struct hashfile_handle *handle) {
  time_t tm = (time_t)handle->metadata.cur_time;
  char *buf = ctime(&tm);
  return buf;
}
void parse_file_dir(char *path) {
  int i = strlen(path);
  if (path[i] != '/') {
    *(path + i) = '/';
    *(path + i + 1) = '\0';
  }
}

char *hashfile_name(char *dir, int len, int hashfile_count) {
  char *file_name = dir;
  sprintf(file_name + len, "%04d", hashfile_count);
  return file_name;
}

int split_line(struct hashfile_handle *handle, char *buffer) {
  /*go to the beginning of the nextline*/
  int ret = 0;
  int i;
  int len = strlen(buffer);
  for (i = 0; i < len; i++) {
    if (buffer[i] == '\r' && buffer[i + 1] == '\n') {
      ret = lseek(handle->fd, 0, SEEK_CUR);
      ret = lseek(handle->fd, (off_t)-len + i + 2, SEEK_CUR);
      break;
    }
  }
  return ret;
}
int get_char_value(struct hashfile_handle *handle, char *value) {
  int ret = 0;
  char buffer[MAX_B] = "";
  ret = read(handle->fd, buffer, sizeof(buffer));
  int i;
  int len = strlen(buffer);
  for (i = 0; i < len; i++) {
    if (buffer[i] == '\r') {
      ret = lseek(handle->fd, (off_t)-len + i + 2, SEEK_CUR);
      break;
    }
    value[i] = buffer[i];
  }
  value[i] = '\0';
  return ret;
}
int get_int_value(struct hashfile_handle *handle, int *value) {
  int ret;
  char buffer[MAX_B] = "";
  ret = read(handle->fd, buffer, sizeof(buffer));
  if (strlen(buffer) > (int)sizeof(buffer)) buffer[MAX_B - 1] = '\0';
  *value = (int)strtol(buffer, NULL, 10);
  ret = split_line(handle, buffer);
  return ret;
}
int get_uint64_value(struct hashfile_handle *handle, uint64_t *value) {
  int ret;
  char buffer[MAX_B] = "";
  ret = read(handle->fd, buffer, sizeof(buffer));
  *value = (uint64_t)strtoull(buffer, NULL, 10);
  ret = split_line(handle, buffer);
  return ret;
}
int get_key_value(struct hashfile_handle *handle, char *key, int key_size,
                  int *value) {
  int ret;
  ret = read(handle->fd, key, key_size);
  ret = lseek(handle->fd, 1, SEEK_CUR);
  ret = get_int_value(handle, value);
  return ret;
}

int Cur_part_end(struct hashfile_handle *handle) {
  int ret;
  int flag_r = 0, flag_rn = 0, flag_rnr = 0;
  int end = 0;
  while (!end) {
    char buffer[MAX_B] = "";
    ret = read(handle->fd, buffer, sizeof(buffer));
    int len = strlen(buffer);
    if (ret == 0) {
      break;
    }
    int i;
    for (i = 0; i < len; i++) {
      if (i == 0) {
        if (flag_r == 1) {
          if (buffer[i] == '\n' && buffer[i + 1] == '\r' &&
              buffer[i + 2] == '\n') {
            flag_r = 0;
            ret = lseek(handle->fd, (off_t)-len + 3, SEEK_CUR);
            end = 1;
            break;
          } else {
            flag_r = 0;
          }
        } else if (flag_rn == 1) {
          if (buffer[i] == '\r' && buffer[i + 1] == '\n') {
            flag_rn = 0;
            ret = lseek(handle->fd, (off_t)-len + 2, SEEK_CUR);
            end = 1;
            break;
          } else {
            flag_rn = 0;
          }
        } else if (flag_rnr == 1) {
          if (buffer[i] == '\n') {
            flag_rnr = 0;
            ret = lseek(handle->fd, (off_t)-len + 1, SEEK_CUR);
            end = 1;
            break;
          } else {
            flag_rnr = 0;
          }
        }
      } else if (i < len - 3 && buffer[i] == '\r' && buffer[i + 1] == '\n' &&
                 buffer[i + 2] == '\r' && buffer[i + 3] == '\n') {
        ret = lseek(handle->fd, (off_t)-len + i + 4, SEEK_CUR);
        end = 1;
        break;
      } else if (i == len - 3 && buffer[i] == '\r' && buffer[i + 1] == '\n' &&
                 buffer[i + 2] == '\r') {
        flag_rnr = 1;
      } else if (i == len - 2 && buffer[i] == '\r' && buffer[i + 1] == '\n') {
        flag_rn = 1;
      } else if (i == len - 1 && buffer[i] == '\r') {
        flag_r = 1;
      }
    }
  }
  return ret;
}
struct hashfile_handle *hashfile_open(char *hashfile_name) {
  int fd;
  int ret;
  struct hashfile_handle *handle;
  int saved_errno = 0;
  char buffer[MAX_B] = "";
  handle = (struct hashfile_handle *)malloc(sizeof(*handle));
  if (!handle) goto out;

  fd = open(hashfile_name, O_RDONLY);
  if (fd < 0) {
    saved_errno = errno;
    goto free_handle;
  }
  handle->fd = fd;
  ret = read(fd, buffer, sizeof(buffer));
  if (ret != sizeof(buffer)) {
    if (ret >= 0)
      saved_errno = EAGAIN;
    else
      saved_errno = errno;
    goto close_file;
  }
  if (buffer[0] == 'S') {
    handle->metadata.scan_method = STANDARD;
  } else if (buffer[0] == 'B') {
    handle->metadata.scan_method = BACKUP;
  }
  ret = split_line(handle, buffer);
  ret = get_char_value(handle, handle->metadata.username);
  ret = get_char_value(handle, handle->metadata.hostname);
  ret = get_char_value(handle, handle->metadata.sys_dir);
  ret = get_uint64_value(handle, &handle->metadata.cur_time);
  ret = get_char_value(handle, handle->metadata.OS);
  handle->metadata.hsh_method = MD5_48BIT_HASH;
  handle->current_file.chunks = 0;
  handle->num_files_processed = 0;
  handle->num_hashes_processed_current_files = 0;
  handle->current_chunk_info.hash = (char *)malloc(sizeof(char) * 12);
  if (!handle->current_chunk_info.hash) {
    saved_errno = errno;
    goto close_file;
  }
  ret = Cur_part_end(handle);
  return handle;

close_file:
  close(fd);
free_handle:
  free(handle);
  errno = saved_errno;
out:
  return NULL;
}

int hashfile_next_file(struct hashfile_handle *handle) {
  char buffer[MAX_B] = "";
  int ret;
  ret = read(handle->fd, buffer, sizeof(char) * 11);
  if (strcmp(buffer, "LOGCOMPLETE") == 0) {
    return 0;
  } else {
    ret = lseek(handle->fd, (off_t)-11, SEEK_CUR);
  }

  ret = get_key_value(handle, handle->current_file.dir_name,
                      sizeof(handle->current_file.dir_name),
                      &handle->current_file.dir_length);
  ret = get_key_value(handle, handle->current_file.file_name,
                      sizeof(handle->current_file.file_name),
                      &handle->current_file.file_length);
  ret = get_key_value(handle, handle->current_file.extensions,
                      sizeof(handle->current_file.extensions),
                      &handle->current_file.ext_value);
  ret = get_int_value(handle, &handle->current_file.namespace_depth);
  ret = get_uint64_value(handle, &handle->current_file.file_size);
  assert(handle->current_file.file_size >= 0);
  ret = get_char_value(handle, handle->current_file.attr_flags);
  ret = get_uint64_value(handle, &handle->current_file.file_id);
  ret = get_int_value(handle, &handle->current_file.hardlinks);
  ret = get_char_value(handle, handle->current_file.reparse_flags);
  ret = get_uint64_value(handle, &handle->current_file.ctime);
  assert(handle->current_file.ctime != 0);
  ret = get_uint64_value(handle, &handle->current_file.atime);
  ret = get_uint64_value(handle, &handle->current_file.mtime);
  return ret;
}

uint64_t hashfile_curfile_size(struct hashfile_handle *handle) {
  return handle->current_file.file_size;
}
int chunk_hash_size(struct chunk_info *ci) { return (int)strlen(ci->hash); }
const struct chunk_info *hashfile_next_chunk(struct hashfile_handle *handle) {
  int ret = 0;
  char buffer[MAX_B] = "";
  if (!hashfile_curfile_size(handle)) { /* file_size == 0*/
    ret = read(handle->fd, buffer, sizeof(buffer));
    if (buffer[0] == '\r' && buffer[1] == '\n') {  // No Lcn, Vcn
      ret = lseek(handle->fd, (off_t)-strlen(buffer) + 2, SEEK_CUR);
    } else {
      ret = lseek(handle->fd, (off_t)-strlen(buffer), SEEK_CUR);
      ret = Cur_part_end(handle);
    }
    return NULL;
  }

  while (1) {
    memset(buffer, 0, sizeof(buffer));
    ret = read(handle->fd, buffer, sizeof(buffer));
    if (buffer[0] == 'S' || buffer[0] == 'V' || buffer[0] == 'A') {
      ret = split_line(handle, buffer);
      continue;
    } else if (buffer[0] == '\r' && buffer[1] == '\n') {
      // handle exception: file_size is not 0, but no chunk
      ret = split_line(handle, buffer);
      return NULL;
    } else {
      ret = lseek(handle->fd, (off_t)-strlen(buffer), SEEK_CUR);
      int hash_size = 10;
      if (buffer[0] == 'z')
        hash_size = 12;
      else
        hash_size = 10;
      ret = get_key_value(handle, handle->current_chunk_info.hash, hash_size,
                          &handle->current_chunk_info.size);
      if (hash_size == 10) {
        handle->current_chunk_info.hash[10] = '\0';
      }
      if (hash_size == 12) {
        handle->current_chunk_info.hash[12] = '\0';
      }
      assert(handle->current_chunk_info.size != 0);
      break;
    }
  }
  return &handle->current_chunk_info;
}
void hashfile_close(struct hashfile_handle *handle) {
  free(handle->current_chunk_info.hash);
  close(handle->fd);
  free(handle);
}
